<?php
namespace App\Service;

class WebhookService {

    const MAX_ACK_EVENTS = 10000;
    
    const MAX_PARALLEL_ORDERS = 120;
    
    private $webhookFactory;
    private $util;

    //vida do token deverá ser verificada no proprio token
    //criar campo para salvar data de expiração do token
    //utilizar biblioteca jwt
    const IFO_TOKEN_LIFE = 600; // ten minutes
            
    public function __construct (\App\Factory\WebhookFactory $webhookFactory,
                                \App\Util\Util $util) {
        $this->webhookFactory = $webhookFactory;
        $this->util = $util;
    }
    
    public function getClientData($token) {
        /*if(strlen($token) != 36) {
           // throw new \Exception("Invalid token (36 char)");            
        }*/
        
        return $this->webhookFactory->getClientData($token);        
    }

    public function getTokensByClient($clientId) {
        
        return $this->webhookFactory->getTokensByClient($clientId);
    }
    
    public function getAllTokens() {
        //return $this->webhookFactory->getAllTokens();
        
        $configFile = file_get_contents(__DIR__ . "/../../environment/config.json");
        $configData = json_decode($configFile, true);
        $basePath   = $configData['basePath'];
        
        $filePath = $basePath.'/activeMerchants.json';
        
        return json_decode(file_get_contents($filePath), true);
    }
    
    public function getClientList() {
        $clientList = $this->webhookFactory->getClientList();

        $i = 0;
        $result = [];

        foreach($clientList as $client) {
            $result[$i] = [
                'idCliente'   => $client['IDCLIENTE'],
                'cliente'     => $client['NOME'],
                'url'         => $client['URL'],
                'tipoCliente' => $client['TIPO'],
                'usuarioDb'   => $client['DBNAME'],
                'version'     => $this->getClientVersion($client['URL']),
                'tokens'      => $this->getTokensByClient($client['IDCLIENTE'])
            ];
            $i++;
        }


        return $result;
    }    

    private function getClientVersion($url) {
        return '1.2.3 - rev 6';
    }

    public function addClient($params) {

        if(!isset($params['nomeCliente']) || empty($params['nomeCliente'])) {
            throw new \Exception("Empty nomeCliente");            
        }

        if(!isset($params['urlCliente']) || empty($params['urlCliente'])) {
            throw new \Exception("Empty urlCliente");            
        }
        
        if(!isset($params['tipoCliente']) || empty($params['tipoCliente'])) {
            throw new \Exception("Empty tipoCliente");            
        }

        if(!isset($params['usuarioDb']) || empty($params['usuarioDb'])) {
            throw new \Exception("Empty usuarioDb");            
        }
        $nomeCliente = trim($params['nomeCliente']);
        $urlCliente  = trim($params['urlCliente']);
        $tipoCliente = trim($params['tipoCliente']);
        $usuarioDb  = trim($params['usuarioDb']);
        $this->webhookFactory->addClient($nomeCliente, $urlCliente, $tipoCliente, $usuarioDb);

        return $this->webhookFactory->getClient($nomeCliente, $urlCliente);
    }

    public function editClient($params) {

        if(!isset($params['idCliente']) || empty($params['idCliente'])) {
            throw new \Exception("Empty idCliente");            
        }

        if(!isset($params['nomeCliente']) || empty($params['nomeCliente'])) {
            throw new \Exception("Empty nomeCliente");            
        }

        if(!isset($params['urlCliente']) || empty($params['urlCliente'])) {
            throw new \Exception("Empty urlCliente");            
        }
        
         if(!isset($params['tipoCliente']) || empty($params['tipoCliente'])) {
            throw new \Exception("Empty tipoCliente");            
        }
        
        if(!isset($params['usuarioDb']) || empty($params['usuarioDb'])) {
            throw new \Exception("Empty usuarioDb");            
        }
        
        $idCliente   = trim($params['idCliente']);
        $nomeCliente = trim($params['nomeCliente']);
        $urlCliente  = trim($params['urlCliente']);
        $tipoCliente = trim($params['tipoCliente']);
        $usuarioDb   = trim($params['usuarioDb']);

        $this->webhookFactory->editClient($idCliente, $nomeCliente, $urlCliente, $tipoCliente, $usuarioDb);

        return $this->webhookFactory->getClient($nomeCliente, $urlCliente);
    }

    public function deleteClient($params) {
        if(!isset($params['idCliente']) || empty($params['idCliente'])) {
            throw new \Exception("Empty idCliente");            
        }

        $idCliente = trim($params['idCliente']);

        if(empty($this->webhookFactory->getClientById($idCliente))) {
            throw new \Exception("Invalid idCliente");            
        }

        $this->webhookFactory->deleteClient($idCliente);
    }

    public function addToken($params) {

        if(!isset($params['idCliente']) || empty($params['idCliente'])) {
            throw new \Exception("Empty idCliente");            
        }

        if(!isset($params['token']) || empty($params['token'])) {
            throw new \Exception("Empty urlCliente");            
        }

        $idCliente = trim($params['idCliente']);
        $token     = trim($params['token']);
        
        if(empty($this->webhookFactory->getClientById($idCliente))) {
            throw new \Exception("Invalid idCliente");            
        }
        
        //validate token
        if(!empty($this->webhookFactory->getTokenByToken($token))) {
            throw new \Exception("O token informado já está associado com algum cliente");            
        }

        $this->webhookFactory->addToken($idCliente, $token);
        return $this->webhookFactory->getToken($idCliente, $token);
    }

    public function editToken($params) {

        if(!isset($params['idToken']) || empty($params['idToken'])) {
            throw new \Exception("Empty idToken");            
        }

        if(!isset($params['token']) || empty($params['token'])) {
            throw new \Exception("Empty token");            
        }

        $idToken = trim($params['idToken']);
        $token   = trim($params['token']);
        
        if(empty($this->webhookFactory->getTokenById($idToken))) {
            throw new \Exception("Invalid idToken");            
        }

        $this->webhookFactory->editToken($idToken, $token);
        return $this->webhookFactory->getTokenById($idToken);
    }

    public function deleteToken($params) {
        if(!isset($params['idToken']) || empty($params['idToken'])) {
            throw new \Exception("Empty idToken");            
        }

        $idToken = trim($params['idToken']);

        if(empty($this->webhookFactory->getTokenById($idToken))) {
            throw new \Exception("Invalid idToken");            
        }

        $this->webhookFactory->deleteToken($idToken);
    }
    
    public function cleanToken() {
        $this->webhookFactory->cleanToken();
    }
    
    public function getWebhookParams(){
       return $this->webhookFactory->getWebhookParams();
    }
    
    public function updateParams($params){
        if(!isset($params['appId']) || empty($params['appId'])) {
            throw new \Exception("Empty appId");            
        }

        if(!isset($params['appSecret']) || empty($params['appSecret'])) {
            throw new \Exception("Empty appSecret");            
        }

        $appId = trim($params['appId']);
        $appSecret   = trim($params['appSecret']);

        $this->webhookFactory->updateParams($appSecret, $appId);
        return $this->webhookFactory->getWebhookParams();
    }
    
    public function getAuthToken() {
        $token = '';
        
        $webhookParams = $this->webhookFactory->getWebhookParams();
        
        $clientId      = $webhookParams['CLIENTID'];
        $clientSercret = $webhookParams['CLIENTSECRET'];
        
        if(empty($clientId) || empty($clientSercret))
            throw new \Exception("Invalid clientId or clientSecret");
        
        if($this->isAuthTokenExpired($webhookParams)) {
            
            $token = json_decode($this->util->authenticate($clientId, $clientSercret), true);
            
            if(empty($token))
                throw new \Exception("token is empty");
            
            $token = $token['accessToken'];
            
            $createdAt = time();
            
            $this->webhookFactory->updateWebhookParams($token, $createdAt);
        } else {
            $token = $webhookParams['CURRENTTOKEN'];
        }
        
        return [
            'token' => $token
        ];
    }

    private function isAuthTokenExpired($webhookParams) {
        if(empty($webhookParams['CURRENTTOKEN']))
            return 1;
        
        if(empty($webhookParams['TKNDTCRIAC']) || ((time() - intval($webhookParams['TKNDTCRIAC'])) > self::IFO_TOKEN_LIFE) )
            return 1;
        
        return 0;
    }

    public function getMerchantsForDispatch() {
        return $this->webhookFactory->getMerchantsForDispatch();
    }

    public function getMerchantStatus($merchantId, $token) {
        $curl = curl_init();
        
        //IFOOD URL
        $url = 'https://merchant-api.ifood.com.br/merchant/v1.0/merchants/';
        
        //TEKNISA SANDBOX API URL
        //$url = 'https://heitorcarvalho1.zeedhi.com/workfolder/odhen-ifood-bench/src/public/index.php/order/v1.0/merchant/';
        
        $header = array(
            "Authorization: bearer $token"
        );
        
        
        curl_setopt_array($curl, array(
            CURLOPT_URL            => $url.$merchantId."/status",
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING       => '',
            CURLOPT_MAXREDIRS      => 10,
            CURLOPT_TIMEOUT        => 10,
            CURLOPT_CONNECTTIMEOUT => 5,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST  => 'GET',
            CURLOPT_HTTPHEADER     => $header
        ));
        
        $response = curl_exec($curl);
        $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE);

        if (!empty(curl_error($curl)) || (stripos($response, "<html>"))) {
            $fileName = date("H-i-s") . ".txt";
            $this->util->writeLog('network', $fileName, $this->util->formatNetworkError([
                'eventTime' => date("d/m/Y H:i:s"),
                'header'    => $header,
                'url'       => $url,
                'method'    => 'GET',
                'params'    => [],
                'httpCode'  => $httpCode,
                'error'     => curl_error($curl),
                'response'  => $response
            ]));
        }
        
        curl_close($curl);
        
        return $response;
    }
    
    public function getMerchantInterruptions($merchantId, $token) {
        $curl = curl_init();
        
        //IFOOD URL
        $url = 'https://merchant-api.ifood.com.br/merchant/v1.0/merchants/';
        
        //TEKNISA SANDBOX API URL
        //$url = 'https://heitorcarvalho1.zeedhi.com/workfolder/odhen-ifood-bench/src/public/index.php/order/v1.0/merchant/';
        
        $header = array(
            "Authorization: bearer $token"
        );
        
        
        curl_setopt_array($curl, array(
            CURLOPT_URL            => $url.$merchantId."/interruptions",
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING       => '',
            CURLOPT_MAXREDIRS      => 10,
            CURLOPT_TIMEOUT        => 10,
            CURLOPT_CONNECTTIMEOUT => 5,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST  => 'GET',
            CURLOPT_HTTPHEADER     => $header
        ));
        
        $response = curl_exec($curl);
        $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE);

        if (!empty(curl_error($curl)) || (stripos($response, "<html>"))) {
            $fileName = date("H-i-s") . ".txt";
            $this->util->writeLog('network', $fileName, $this->util->formatNetworkError([
                'eventTime' => date("d/m/Y H:i:s"),
                'header'    => $header,
                'url'       => $url,
                'method'    => 'GET',
                'params'    => [],
                'httpCode'  => $httpCode,
                'error'     => curl_error($curl),
                'response'  => $response
            ]));
        }
        
        curl_close($curl);
        
        return $response;
    }
    
    public function requestCancelOrder($orderId, $token) {
        $curl = curl_init();
        
        //IFOOD URL
        $url = 'https://merchant-api.ifood.com.br/order/v1.0/orders/';
        
        //TEKNISA SANDBOX API URL
        //$url = 'https://heitorcarvalho1.zeedhi.com/workfolder/odhen-ifood-bench/src/public/index.php/order/v1.0/merchant/';
        
        $header = array(
            "Content-Type: application/json",
            "Authorization: bearer $token"
        );
        
        $body = [
            "reason" => "PEDIDO CANCELADO PELO RESTAURANTE",
            "cancellationCode" => "501"
        ];
        
        curl_setopt_array($curl, array(
            CURLOPT_URL            => $url.$orderId."/requestCancellation",
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING       => '',
            CURLOPT_MAXREDIRS      => 10,
            CURLOPT_TIMEOUT        => 10,
            CURLOPT_CONNECTTIMEOUT => 5,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST  => 'POST',
            CURLOPT_POSTFIELDS     => json_encode($body),
            CURLOPT_HTTPHEADER     => $header
        ));
        
        $response = curl_exec($curl);
        $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE);

        if (!empty(curl_error($curl)) || (stripos($response, "<html>"))) {
            $fileName = date("H-i-s") . ".txt";
            $this->util->writeLog('network', $fileName, $this->util->formatNetworkError([
                'eventTime' => date("d/m/Y H:i:s"),
                'header'    => $header,
                'url'       => $url,
                'method'    => 'GET',
                'params'    => [],
                'httpCode'  => $httpCode,
                'error'     => curl_error($curl),
                'response'  => $response
            ]));
        }
        
        curl_close($curl);
        
        return $httpCode;
    }
    
    public function getEvents($token, $arrayMerchantList) {
        //IFOOD URL
        $url = 'https://merchant-api.ifood.com.br/order/v1.0/events:polling';
        
        //TEKNISA SANDBOX API URL
        //$url = 'https://heitorcarvalho1.zeedhi.com/workfolder/odhen-ifood-bench/src/public/index.php/order/v1.0/events:polling';

        $running = null;
        $handles = [];
        $c = 0;

        $numberOfPollings = 0;

        try {
            $numberOfPollings = count($arrayMerchantList);
            $headerArray = [];

            if($numberOfPollings > 0) {
                
                foreach($arrayMerchantList as $merchantList) {
                    
                    $header = array(
                        "Authorization: bearer $token",
                        "x-polling-merchants: $merchantList"
                    );
                    
                    $headerArray[$c] = $header;
                    
                    $handles[$c] = curl_init();

                    curl_setopt_array($handles[$c], array(
                        CURLOPT_URL            => $url,
                        CURLOPT_RETURNTRANSFER => true,
                        CURLOPT_ENCODING       => '',
                        CURLOPT_MAXREDIRS      => 10,
                        CURLOPT_TIMEOUT        => 10,
                        CURLOPT_CONNECTTIMEOUT => 5,
                        CURLOPT_FOLLOWLOCATION => true,
                        CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
                        CURLOPT_CUSTOMREQUEST  => 'GET',
                        CURLOPT_HTTPHEADER     => $header
                    ));

                    $c++;

                    if($c >= self::MAX_PARALLEL_ORDERS) {
                        break;
                    }
                }

                $mh = curl_multi_init();
                for($i = 0; $i < $c; $i++) {
                    curl_multi_add_handle($mh, $handles[$i]);
                }
                    
                do {
                    curl_multi_exec($mh, $running);
                } while ($running > 0);
                
                for($i = 0; $i < $c; $i++){
                    $info = curl_multi_getcontent($handles[$i]);
                    $requestInfo = curl_getinfo($handles[$i]);
                    $url = $requestInfo['url'];
                    $httpCode = $requestInfo['http_code'];
                    
                    if (($httpCode !== 200 && $httpCode !== 201 && $httpCode !== 202 && $httpCode !== 204) || (stripos($info, "<html>"))){
                        if($httpCode === 0){
                            $info = 'Timeout was reached.';
                        }
                        
                        $requestData = [
                            'header'    => $headerArray[$i],
                            'url'       => $url,
                            'method'    => 'GET',
                            'params'    => [],
                            'httpCode'  => $httpCode,
                            'error'     => $info
                        ];
                        
                        $fileName = date("H-i-s") . "-" . round(microtime(true) * 1000) . ".txt";
                        
                        $this->util->writeLog('network', $fileName, $this->util->formatMultiNetworkError($requestData));
                    } else{
                        $pollingData = json_decode($info, true);
                        if(!empty($pollingData)){
                            $this->saveEvents($pollingData);   
                        }
                    }
                }

                for($i = 0; $i < $c; $i++) {
                    curl_multi_remove_handle($mh, $handles[$i]);
                }

                curl_multi_close($mh);
            }
        } catch(\Exception $e) {
            throw new \Exception($e->getMessage());
        }
    }
    
    public function saveEvents($events) {
        foreach($events as $event) {
            //salvar evento em arquivo - OK
            //so deve salvar o evento na pasta pending e received caso ele não exista no diretório ack - OK
            //caso o evento venha repetido dentro do polling o arquivo será sobreescrito evitando problemas de duplicidade - OK

            $fileName = $event['id'].".json";
            $receivedFileName = $event['orderId'].".json";

            $this->util->manipulateEventFile("pending", $fileName, json_encode($event), $receivedFileName);
        }
    }
    
    public function runPendingACKs($token) {
        //ler eventos dos arquivos da pasta pending - OK
        $pendingAckEvents = $this->util->readEventFiles("pending");

        $ackEvents      = [];
        $events         = [];
        $cancelEvents   = [];
        $cancReqEvents  = [];
        $cancFailEvents = [];

        foreach($pendingAckEvents as $event) {
            if(count($ackEvents) >= self::MAX_ACK_EVENTS) {
                break;
            }
            
            if ($event["fullCode"] === "PLACED") {
                $this->verifyEventDup($event, $events);
            } else if ($event["fullCode"] === "CANCELLED") {
                $this->verifyEventDup($event, $cancelEvents);
            } else if ($event["fullCode"] === "CANCELLATION_REQUESTED") {
                $this->verifyEventDup($event, $cancReqEvents);
            } else if ($event["fullCode"] === "CANCELLATION_REQUEST_FAILED") {
                $this->verifyEventDup($event, $cancFailEvents);
            }

            array_push($ackEvents, $event);
        }

        $this->filterEventsArray($cancelEvents, $events);
        $this->filterEventsArray($cancReqEvents, $events);
        $this->filterEventsArray($cancFailEvents, $events);
        
        if(!empty($ackEvents)) {
            $this->acknowledgment($ackEvents, $token, $events, $cancelEvents, $cancReqEvents);
        }
    }

    private function verifyEventDup($unfilteredEvent, &$eventsArray) {
        foreach($eventsArray as $event) {
            if($event['orderId'] === $unfilteredEvent['orderId']) {
                return 0;
            }
        }
        array_push($eventsArray, $unfilteredEvent);
    }
    
    private function filterEventsArray($ordersEvents, &$events) {
        foreach($ordersEvents as $ordersEvent) {
            foreach($events as &$event) {
                if ($ordersEvent['orderId'] === $event['orderId'])
                    array_splice($events, array_search($event, $events), 1);
            }
            if(isset($event))
                unset($event);
        }
    }
    
    public function acknowledgment($ackEvents, $token, $events, $cancelEvents, $cancReqEvents) {
        
        $curl = curl_init();
        
        //IFOOD URL
        $url = 'https://merchant-api.ifood.com.br/order/v1.0/events/acknowledgment';
        
        //TEKNISA SANDBOX API URL
        //$url = 'https://heitorcarvalho1.zeedhi.com/workfolder/odhen-ifood-bench/src/public/index.php/order/v1.0/events/acknowledgment';
        $header = array(
            "Content-Type: application/json",
            "Authorization: bearer $token"
        );
        
        $body = [];

        foreach ($ackEvents as $event) {
            array_push($body, [
                "id" => $event["id"]
            ]);
        }
        
        curl_setopt_array($curl, array(
            CURLOPT_URL            => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING       => '',
            CURLOPT_MAXREDIRS      => 10,
            CURLOPT_TIMEOUT        => 15,
            CURLOPT_CONNECTTIMEOUT => 5,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST  => 'POST',
            CURLOPT_POSTFIELDS     => json_encode($body),
            CURLOPT_HTTPHEADER     => $header
        ));
        
        $response = curl_exec($curl);
        $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE);

        if (!empty(curl_error($curl)) || (stripos($response, "<html>"))) {
            $fileName = date("H-i-s") . ".txt";
            $this->util->writeLog('network', $fileName, $this->util->formatNetworkError([
                'eventTime' => date("d/m/Y H:i:s"),
                'header'    => $header,
                'url'       => $url,
                'method'    => 'POST',
                'params'    => json_encode($body),
                'httpCode'  => $httpCode,
                'error'     => curl_error($curl),
                'response'  => $response
            ]));
        } else {
            //mover arquivos com ack bem sucedido da pasta pending para a pasta ack - OK
            //apos o processamento do pedido, os arquivos serão copiados da pasta ack para a pasta processed - Feito durante a integracao do pedido
            //so deverá ser feito o processamento de arquivos que estão na pasta ack e não estão na pasta processed - Feito durante a integracao do pedido
            //arquivos serão mantidos na pasta ack por um tempo determinado apos o processamento para fins de controle - Feito durante a integracao do pedido
            
            foreach($events as $event) {
                $fileName = $event['id'].".json";
                $this->util->manipulateEventFile("order", $fileName, json_encode($event));
            }

            foreach($cancelEvents as $cancelEvent) {
                $fileName = $cancelEvent['id'].".json";
                $this->util->manipulateEventFile("cancelled", $fileName, json_encode($cancelEvent));
            }

            foreach($cancReqEvents as $cancReqEvent) {
                $fileName = $cancReqEvent['id'].".json";
                $this->util->manipulateEventFile("cancelRequest", $fileName, json_encode($cancReqEvent));
            }

            foreach($ackEvents as $ackEvent) {
                $fileName = $ackEvent['id'].".json";
                $this->util->manipulateEventFile("ack", $fileName, json_encode($ackEvent));
            }
        }
        
        curl_close($curl);
    }

}
